/* Config */
#include "FlashConfig.h"

/* HAL */
#include "py32f0xx.h"

/* Private */
#include "FlashOperations.h"

#define APP_HSI_CALIB_24M_ADDR (0x1FFF0110UL)
#define APP_HSI_CALIB_VAL      (*(uint8_t *)APP_HSI_CALIB_24M_ADDR)

#define APP_FLASH_CALIB_24M_ADDR (0x1FFF0F6CUL)
#define APP_FLASH_KEY1           (0x45670123UL)
#define APP_FLASH_KEY2           (0xCDEF89ABUL)

static inline int RCC_Init(void) {
    /* Calibrate HSI */
    RCC->ICSCR = (RCC->ICSCR & ~RCC_ICSCR_HSI_FS_Msk) | (4U << RCC_ICSCR_HSI_FS_Pos);
    RCC->ICSCR = (RCC->ICSCR & ~RCC_ICSCR_HSI_TRIM_Msk) | (APP_HSI_CALIB_VAL << RCC_ICSCR_HSI_TRIM_Pos);

    /* Enable HSI */
    RCC->CR |= RCC_CR_HSION_Msk;
    while ((RCC->CR & RCC_CR_HSIRDY_Msk) == 0UL) {
        /* -- Wait for HSI ready -- */
    }

    /* HSI DIV : /1 */
    RCC->CR = (RCC->CR & ~RCC_CR_HSIDIV_Msk) | (0U << RCC_CR_HSIDIV_Pos);

    /* Switch System Clock to HSI */
    RCC->CFGR = (RCC->CFGR & ~(RCC_CFGR_SW_Msk)) | (0U << RCC_CFGR_SW_Pos);
    while ((RCC->CFGR & RCC_CFGR_SWS_Msk) != (0U << RCC_CFGR_SWS_Pos)) {
        /* -- Wait for switch complete -- */
    }

    /* AHB DIV: /1 */
    RCC->CFGR = (RCC->CFGR & ~(RCC_CFGR_HPRE_Msk)) | (0U << RCC_CFGR_HPRE_Pos);

    return 0;
}

static inline int FMU_Init(void) {
    /* Calibrate FMU */
    FLASH->TS0     = ((*(volatile uint32_t *)APP_FLASH_CALIB_24M_ADDR + 0x00U) >> 0U) & 0xFFU;
    FLASH->TS1     = ((*(volatile uint32_t *)APP_FLASH_CALIB_24M_ADDR + 0x00U) >> 16U) & 0x1FFU;
    FLASH->TS2P    = ((*(volatile uint32_t *)APP_FLASH_CALIB_24M_ADDR + 0x04U) >> 0U) & 0xFFU;
    FLASH->TS3     = ((*(volatile uint32_t *)APP_FLASH_CALIB_24M_ADDR + 0x00U) >> 8U) & 0xFFU;
    FLASH->TPS3    = ((*(volatile uint32_t *)APP_FLASH_CALIB_24M_ADDR + 0x04U) >> 16U) & 0x7FFU;
    FLASH->PERTPE  = ((*(volatile uint32_t *)APP_FLASH_CALIB_24M_ADDR + 0x08U) >> 0U) & 0x1FFFFU;
    FLASH->SMERTPE = ((*(volatile uint32_t *)APP_FLASH_CALIB_24M_ADDR + 0x0CU) >> 0U) & 0x1FFFFU;
    FLASH->PRGTPE  = ((*(volatile uint32_t *)APP_FLASH_CALIB_24M_ADDR + 0x10U) >> 0U) & 0xFFFFU;
    FLASH->PRETPE  = ((*(volatile uint32_t *)APP_FLASH_CALIB_24M_ADDR + 0x10U) >> 16U) & 0xFFFFU;
    return 0;
}

static inline int Flash_CheckBusy(void) {
    if (FLASH->SR & FLASH_SR_BSY_Msk) {
        return -1;
    }
    return 0;
}

static int Flash_CheckEoP(void) {
    while (FLASH->SR & FLASH_SR_BSY_Msk) {
        /* -- */
    }

    /* Check end-of-process */
    if ((FLASH->SR & FLASH_SR_EOP_Msk) == 0U) {
        return -1;
    }

    FLASH->SR |= FLASH_SR_EOP_Msk;

    return 0;
}

static inline int Flash_Unlock(void) {
    /* Unlock CR register */
    if (FLASH->CR & FLASH_CR_LOCK_Msk) {
        FLASH->KEYR = APP_FLASH_KEY1;
        FLASH->KEYR = APP_FLASH_KEY2;
    }

    /* Check unlock result */
    if (FLASH->CR & FLASH_CR_LOCK_Msk) {
        return -1;
    }

    return 0;
}

static inline void Flash_Lock(void) {
    FLASH->CR |= FLASH_CR_LOCK_Msk;
}

int Flash_Init(void) {
    RCC_Init();
    FMU_Init();
    return Flash_Unlock();
}

int Flash_EraseSector(uint32_t addr) {
    /* Check FMU busy status */
    if (Flash_CheckBusy() != 0) {
        return -1;
    }

    /* Select sector erase mode, de-select all other modes */
    FLASH->CR &= ~(FLASH_CR_PG_Msk | FLASH_CR_MER_Msk | FLASH_CR_PER_Msk);
    FLASH->CR |= FLASH_CR_SER_Msk | FLASH_CR_EOPIE_Msk;

    /* Write arbitrary data to the sector to begin erase operation */
    *(volatile uint32_t *)addr = 0xFFFFFFFFUL;

    if (Flash_CheckEoP() != 0) {
        return -3;
    }

    return 0;
}

int Flash_ProgramPage(uint32_t addr, uint8_t *data, uint32_t len) {
    /* Check FMU busy status */
    if (Flash_CheckBusy() != 0) {
        return -1;
    }

    /* Select page program mode, de-select all other modes */
    FLASH->CR &= ~(FLASH_CR_SER_Msk | FLASH_CR_MER_Msk | FLASH_CR_PER_Msk);
    FLASH->CR |= FLASH_CR_PG_Msk | FLASH_CR_EOPIE_Msk;

    for (uint8_t i = 0U; i < 32; i++) {
        if (i == 31) {
            FLASH->CR |= FLASH_CR_PGSTRT_Msk;
        }

        *(volatile uint32_t *)(addr + i * 4) = ((uint32_t *)data)[i];
    }

    if (Flash_CheckEoP() != 0) {
        return -3;
    }

    return 0;
}

int Flash_EraseChip(void) {
    /* Check FMU busy status */
    if (Flash_CheckBusy() != 0) {
        return -1;
    }

    /* Select mass erase mode, de-select all other modes */
    FLASH->CR &= ~(FLASH_CR_PG_Msk | FLASH_CR_SER_Msk | FLASH_CR_PER_Msk);
    FLASH->CR |= FLASH_CR_MER_Msk | FLASH_CR_EOPIE_Msk;

    /* Write arbitrary data to flash to begin erase operation */
    *(uint32_t *)APP_FLASH_BASE = 0xFFFFFFFFUL;

    if (Flash_CheckEoP() != 0) {
        return -2;
    }

    return 0;
}

int Flash_Deinit(void) {
    Flash_Lock();
    return 0;
}